# -*- coding: utf-8 -*-
"""Workchain to run a pyscf calculation with automated error handling and restarts."""
from aiida.common import AttributeDict
from aiida.engine import BaseRestartWorkChain, ProcessHandlerReport, process_handler, while_
from aiida.plugins import CalculationFactory

PyscfCalculation = CalculationFactory('pyscf.base')


class PyscfBaseWorkChain(BaseRestartWorkChain):
    """Workchain to run a pyscf calculation with automated error handling and restarts."""

    _process_class = PyscfCalculation  # type: ignore[assignment]

    @classmethod
    def define(cls, spec):
        """Define the process specification."""
        super().define(spec)
        spec.expose_inputs(PyscfCalculation, namespace='pyscf')
        spec.expose_outputs(PyscfCalculation)
        spec.outline(
            cls.setup,
            while_(cls.should_run_process)(  # type: ignore[arg-type]
                cls.run_process,  # type: ignore[arg-type]
                cls.inspect_process,  # type: ignore[arg-type]
            ),
            cls.results,
        )
        spec.exit_code(
            300, 'ERROR_UNRECOVERABLE_FAILURE', message='The calculation failed with an unrecoverable error.'
        )
        spec.exit_code(
            310,
            'ERROR_NO_CHECKPOINT_TO_RESTART',
            message='The calculation failed and did not retrieve a checkpoint file from which can be restarted.',
        )

    def setup(self):
        """Call the `setup` of the `BaseRestartWorkChain` and then create the inputs dictionary in `self.ctx.inputs`.

        This `self.ctx.inputs` dictionary will be used by the `BaseRestartWorkChain` to submit the calculations in the
        internal loop.
        """
        super().setup()
        self.ctx.restart_calc = None
        self.ctx.inputs = AttributeDict(self.exposed_inputs(PyscfCalculation, 'pyscf'))  # type: ignore[arg-type]

    def report_error_handled(self, calculation, action):
        """Report an action taken for a calculation that has failed.

        This should be called in a registered error handler if its condition is met and an action was taken.

        :param calculation: the failed calculation node
        :param action: a string message with the action taken
        """
        arguments = [calculation.process_label, calculation.pk, calculation.exit_status, calculation.exit_message]
        self.report('{}<{}> failed with exit status {}: {}'.format(*arguments))
        self.report(f'Action taken: {action}')

    @process_handler(priority=600)
    def handle_unrecoverable_failure(self, node):
        """Handle calculations with an exit status below 400 which are unrecoverable, so abort the work chain."""
        if node.is_failed and node.exit_status < 400:
            self.report_error_handled(node, 'unrecoverable error, aborting...')
            return ProcessHandlerReport(True, self.exit_codes.ERROR_UNRECOVERABLE_FAILURE)

    @process_handler(
        priority=500,
        exit_codes=[
            PyscfCalculation.exit_codes.ERROR_IONIC_CONVERGENCE_NOT_REACHED,  # type: ignore[union-attr]
        ],
    )
    def handle_ionic_convergence_not_reached(self, node):
        """Handle ``ERROR_IONIC_CONVERGENCE_NOT_REACHED`` error.

        Simply restart the calculation using the ``checkpoint`` and ``structure`` outputs of the failed calculation as
        starting point.
        """
        self.ctx.inputs.checkpoint = node.outputs.checkpoint

        if 'structure' in node.outputs:
            self.ctx.inputs.structure = node.outputs.structure
        elif 'trajectory' in node.outputs:
            structure = node.outputs.trajectory.get_step_structure(index=-1)
            structure.pbc = self.ctx.inputs.structure.pbc
            structure.store()
            self.ctx.inputs.structure = structure
        else:
            self.logger.warning('no output `structure` or `trajectory`: restarting from input structure.')

        self.report_error_handled(node, 'restarting from the last checkpoint and structure.')
        return ProcessHandlerReport(True)

    @process_handler(
        priority=410,
        exit_codes=[
            PyscfCalculation.exit_codes.ERROR_ELECTRONIC_CONVERGENCE_NOT_REACHED,  # type: ignore[union-attr]
        ],
    )
    def handle_electronic_convergence_not_reached(self, node):
        """Handle ``ERROR_ELECTRONIC_CONVERGENCE_NOT_REACHED`` error.

        Simply restart the calculation using the ``checkpoint`` of the failed calculation as starting point.
        """
        self.ctx.inputs.checkpoint = node.outputs.checkpoint
        self.report_error_handled(node, 'restarting from the last checkpoint.')
        return ProcessHandlerReport(True)

    @process_handler(
        priority=110,
        exit_codes=[
            PyscfCalculation.exit_codes.ERROR_SCHEDULER_NODE_FAILURE,  # type: ignore[union-attr]
        ],
    )
    def handle_scheduler_node_failure(self, node):
        """Handle ``ERROR_SCHEDULER_NODE_FAILURE`` error.

        Simply restart the calculation using the last available ``checkpoint``, which is the output checkpoint of the
        failed calculation if available, or the previous checkpoint that was used in the inputs.
        """
        if 'checkpoint' in node.outputs:
            self.ctx.inputs.checkpoint = node.outputs.checkpoint

        self.report_error_handled(node, 'restarting from the last checkpoint.')
        return ProcessHandlerReport(True)

    @process_handler(
        priority=100,
        exit_codes=[
            PyscfCalculation.exit_codes.ERROR_SCHEDULER_OUT_OF_WALLTIME,  # type: ignore[union-attr]
        ],
    )
    def handle_out_of_walltime(self, node):
        """Handle ``ERROR_SCHEDULER_OUT_OF_WALLTIME`` error.

        Simply restart the calculation using the ``checkpoint`` if it was retrieved, otherwise abort since having to
        restart from scratch with the same walltime is likely going to fail again.
        """
        if 'checkpoint' not in node.outputs:
            self.report_error_handled(node, 'aborting because the failed calculation does not provide a checkpoint.')
            return ProcessHandlerReport(True, self.exit_codes.ERROR_NO_CHECKPOINT_TO_RESTART)

        self.ctx.inputs.checkpoint = node.outputs.checkpoint
        self.report_error_handled(node, 'restarting from the last checkpoint.')
        return ProcessHandlerReport(True)



# -*- coding: utf-8 -*-
"""``CalcJob`` plugin for PySCF."""
from __future__ import annotations

import copy
import io
import numbers
import pathlib
import typing as t

import numpy as np
from aiida.common.datastructures import CalcInfo, CodeInfo
from aiida.common.folders import Folder
from aiida.engine import CalcJob, CalcJobProcessSpec
from aiida.orm import ArrayData, Dict, SinglefileData, StructureData, TrajectoryData
from aiida_shell.data import PickledData
from ase.io.xyz import write_xyz
from jinja2 import Environment, PackageLoader, PrefixLoader
from plumpy.utils import AttributesFrozendict

__all__ = ('PyscfCalculation',)


class PyscfCalculation(CalcJob):
    """``CalcJob`` plugin for PySCF."""

    FILENAME_SCRIPT: str = 'script.py'
    FILENAME_STDOUT: str = 'aiida.out'
    FILENAME_RESULTS: str = 'results.json'
    FILENAME_MODEL: str = 'model.pickle'
    FILENAME_CHECKPOINT: str = 'checkpoint.chk'
    FILENAME_RESTART: str = 'restart.chk'
    FILEPATH_LOG_INI: pathlib.Path = pathlib.Path(__file__).parent / 'templates' / 'geometric_log.ini'
    MAIN_TEMPLATE: str = 'pyscf/script.py.j2'

    @classmethod
    def define(cls, spec: CalcJobProcessSpec):  # type: ignore[override]
        """Define the process specification.

        :param spec: The object to use to build up the process specification.
        """
        super().define(spec)
        spec.input(
            'structure',
            valid_type=StructureData,
            required=True,
            help='Input structure with molecular structure definition.',
        )
        spec.input(
            'parameters',
            valid_type=Dict,
            validator=cls.validate_parameters,
            required=False,
            help='Input parameters used to render the PySCF script template.',
        )
        spec.input(
            'checkpoint',
            valid_type=SinglefileData,
            required=False,
            help='Checkpoint of a previously completed calculation that failed to converge.',
        )
        spec.inputs['code'].required = True

        options = spec.inputs['metadata']['options']  # type: ignore[index]
        options['parser_name'].default = 'pyscf.base'  # type: ignore[index]
        options['resources'].default = {'num_machines': 1, 'tot_num_mpiprocs': 1}  # type: ignore[index]

        spec.output(
            'parameters',
            valid_type=Dict,
            required=False,
            help='Various computed properties parsed from the `FILENAME_RESULTS` output file.',
        )
        spec.output(
            'structure',
            valid_type=StructureData,
            required=False,
            help='The optimized structure if the input parameters contained the `optimizer` key.',
        )
        spec.output(
            'trajectory',
            valid_type=TrajectoryData,
            required=False,
            help='The geometry optimization trajectory if the input parameters contained the `optimizer` key.',
        )
        spec.output(
            'checkpoint',
            valid_type=SinglefileData,
            required=False,
            help='The checkpoint file in case the calculation did not converge. Can be used as an input for a restart.',
        )
        spec.output(
            'model',
            valid_type=PickledData,
            required=False,
            help='The model in serialized form. Can be deserialized and used without having to run the kernel again.',
        )
        spec.output_namespace('fcidump', valid_type=SinglefileData, required=False, help='Computed fcidump files.')
        spec.output_namespace(
            'cubegen.orbitals', valid_type=SinglefileData, required=False, help='Molecular orbitals in `.cube` format.'
        )
        spec.output(
            'cubegen.density', valid_type=SinglefileData, required=False, help='The charge density in `.cube` format.'
        )
        spec.output(
            'cubegen.mep',
            valid_type=SinglefileData,
            required=False,
            help='The molecular electrostatic potential (MEP) in `.cube` format.',
        )
        spec.outputs['cubegen'].required = False
        spec.output(
            'hessian',
            valid_type=ArrayData,
            required=False,
            help='The computed Hessian.',
        )

        spec.exit_code(302, 'ERROR_OUTPUT_STDOUT_MISSING', message='The stdout output file was not retrieved.')
        spec.exit_code(303, 'ERROR_OUTPUT_RESULTS_MISSING', message='The results JSON file was not retrieved.')
        spec.exit_code(
            410,
            'ERROR_ELECTRONIC_CONVERGENCE_NOT_REACHED',
            message='The electronic minimization cycle did not reach self-consistency.',
        )
        spec.exit_code(
            500,
            'ERROR_IONIC_CONVERGENCE_NOT_REACHED',
            message='The ionic minimization cycle did not converge for the given thresholds.',
        )

    @classmethod
    def validate_parameters(cls, value: Dict | None, _) -> str | None:  # noqa: PLR0911, PLR0912
        """Validate the parameters input."""
        if not value:
            return None

        parameters = copy.deepcopy(value.get_dict())

        mean_field = parameters.pop('mean_field', {})
        mean_field_method = mean_field.pop('method', None)
        valid_methods = ['RKS', 'RHF', 'DKS', 'DHF', 'GKS', 'GHF', 'HF', 'KS', 'ROHF', 'ROKS', 'UKS', 'UHF']
        options = ' '.join(valid_methods)

        if mean_field_method is None:
            return f'The `mean_field.method` has to be specified in the `parameters` input, choose from: {options}'

        if mean_field_method not in valid_methods:
            return f'Specified mean field method {mean_field_method} is not supported, choose from: {options}'

        if 'chkfile' in mean_field:
            return (
                'The `chkfile` cannot be specified in the `mean_field` parameters. It is set automatically by the '
                'plugin if the `checkpoint` input is provided.'
            )

        if (localize_orbitals := parameters.pop('localize_orbitals', None)) is not None:
            valid_lo = ('boys', 'cholesky', 'edmiston', 'iao', 'ibo', 'lowdin', 'nao', 'orth', 'pipek', 'vvo')
            method = localize_orbitals.get('method')
            if method is None:
                return f'No method specified in `localize_orbitals` parameters. Choose from: {valid_lo}'

            if method.lower() not in valid_lo:
                return f'Invalid method `{method}` specified in `localize_orbitals` parameters. Choose from: {valid_lo}'

        if (optimizer := parameters.pop('optimizer', None)) is not None:
            valid_solvers = ('geometric', 'berny')
            solver = optimizer.get('solver')

            if solver is None:
                return f'No solver specified in `optimizer` parameters. Choose from: {valid_solvers}'

            if solver.lower() not in valid_solvers:
                return f'Invalid solver `{solver}` specified in `optimizer` parameters. Choose from: {valid_solvers}'

        if (cubegen := parameters.pop('cubegen', None)) is not None:
            orbitals = cubegen.get('orbitals')
            indices = orbitals.get('indices') if orbitals is not None else None

            if orbitals is not None and indices is None:
                return (
                    'If the `cubegen.orbitals` key is specified, the `cubegen.orbitals.indices` key has to be defined '
                    'with a list of indices.'
                )

            if indices is not None and (not isinstance(indices, list) or any(not isinstance(e, int) for e in indices)):
                return f'The `cubegen.orbitals.indices` parameter should be a list of integers, but got: {indices}'

        if (fcidump := parameters.pop('fcidump', None)) is not None:
            active_spaces = fcidump.get('active_spaces')
            occupations = fcidump.get('occupations')
            arrays = []

            for key, data in (('active_spaces', active_spaces), ('occupations', occupations)):
                try:
                    array = np.array(data)
                except ValueError:
                    return f'The `fcipdump.{key}` should be a nested list of integers, but got: {data}'

                arrays.append(array)

                if len(array.shape) != 2 or not issubclass(array.dtype.type, numbers.Integral):
                    return f'The `fcipdump.{key}` should be a nested list of integers, but got: {data}'

            if arrays[0].shape != arrays[1].shape:
                return 'The `fcipdump.active_spaces` and `fcipdump.occupations` arrays have different shapes.'

        # Remove other known arguments
        for key in ('hessian', 'results', 'structure'):
            parameters.pop(key, None)

        if unknown_keys := list(parameters.keys()):
            return f'The following arguments are not supported: {", ".join(unknown_keys)}'

    def get_template_environment(self) -> Environment:
        """Return the template environment that should be used for rendering.

        :returns: The :class:`jinja2.Environment` to be used for rending the input script template.
        """
        environment = Environment(loader=PrefixLoader({'pyscf': PackageLoader('aiida_pyscf.calculations.base')}))
        environment.trim_blocks = True
        environment.lstrip_blocks = True
        environment.keep_trailing_newline = True
        environment.filters['render_python'] = self.filter_render_python
        return environment

    @property
    def inputs(self) -> AttributesFrozendict:
        """Return the inputs attribute dictionary or an empty one.

        This overrides the property of the base class because that can also return ``None``. This override ensures
        calling functions that they will always get an instance of ``AttributesFrozenDict``.
        """
        return super().inputs or AttributesFrozendict()

    def get_parameters(self) -> dict[str, t.Any]:
        """Return the parameters to use for renderning the input script.

        The base dictionary is formed by the ``parameters`` input node, which is supplemented by default values and
        values that are based off other inputs, such as the ``structure`` converted to XYZ format.

        :returns: Complete dictionary of parameters to render the input script.
        """
        if 'parameters' in self.inputs:
            parameters = self.inputs.parameters.get_dict()
        else:
            parameters = {}

        parameters.setdefault('structure', {})['xyz'] = self.prepare_structure_xyz()
        parameters.setdefault('mean_field', {})
        parameters.setdefault('results', {})['filename_output'] = self.FILENAME_RESULTS
        parameters.setdefault('results', {})['filename_model'] = self.FILENAME_MODEL
        parameters.setdefault('results', {}).setdefault('pickle_model', True)

        if 'optimizer' in parameters:
            parameters['optimizer'].setdefault('convergence_parameters', {})['logIni'] = 'log.ini'

        parameters['mean_field']['chkfile'] = self.FILENAME_CHECKPOINT

        if 'checkpoint' in self.inputs:
            parameters['mean_field']['checkpoint'] = self.FILENAME_RESTART

        return parameters

    @classmethod
    def filter_render_python(cls, value: t.Any) -> t.Any:
        """Render the ``value`` in a template literally as it would be rendered in a Python script.

        For now, it simply ensures that string variables are rendered with quotes as the default behavior of Jinja is to
        strip these strings, since it is mostly intended to generate HTML markup and not Python code.
        This filter can be registered in a template environment and then used as:

            {{ variable|render_python }}

        :param value: Any value that needs to be rendered literally.
        :return: The rendered value
        """
        if isinstance(value, str):
            return f"'{value}'"
        return value

    def render_script(self) -> str:
        """Return the rendered input script.

        :returns: The input script template rendered with the parameters provided by ``get_parameters``.
        """
        parameters = self.get_parameters()
        environment = self.get_template_environment()
        return environment.get_template(self.MAIN_TEMPLATE).render(**parameters)

    def prepare_structure_xyz(self) -> str:
        """Return the input structure in XYZ format without the two leading header lines."""
        stream = io.StringIO()
        write_xyz(stream, [self.inputs.structure.get_ase()])
        stream.seek(0)
        return '\n'.join(stream.read().split('\n')[2:])

    def prepare_for_submission(self, folder: Folder) -> CalcInfo:
        """Prepare the calculation for submission.

        :param folder: A temporary folder on the local file system.
        :returns: A :class:`aiida.common.datastructures.CalcInfo` instance.
        """
        script = self.render_script()
        parameters = self.get_parameters()

        with folder.open(self.FILENAME_SCRIPT, 'w') as handle:
            handle.write(script)

        codeinfo = CodeInfo()
        codeinfo.cmdline_params = [self.FILENAME_SCRIPT]
        codeinfo.code_uuid = self.inputs.code.uuid
        codeinfo.stdout_name = self.FILENAME_STDOUT

        calcinfo = CalcInfo()
        calcinfo.codes_info = [codeinfo]
        calcinfo.provenance_exclude_list = []
        calcinfo.retrieve_temporary_list = [self.FILENAME_CHECKPOINT]
        calcinfo.retrieve_list = [
            self.FILENAME_RESULTS,
            self.FILENAME_MODEL,
            self.FILENAME_STDOUT,
        ]

        if 'checkpoint' in self.inputs:
            with self.inputs.checkpoint.open(mode='rb') as handle:
                folder.create_file_from_filelike(handle, self.FILENAME_RESTART)
            calcinfo.provenance_exclude_list.append(self.FILENAME_RESTART)

        if 'hessian' in parameters:
            calcinfo.retrieve_temporary_list.append('hessian.npy')

        if 'cubegen' in parameters:
            calcinfo.retrieve_temporary_list.append('*.cube')

        if 'fcidump' in parameters:
            calcinfo.retrieve_temporary_list.append('*.fcidump')

        if 'optimizer' in parameters:
            calcinfo.retrieve_temporary_list.append('*_optim.xyz')

            with self.FILEPATH_LOG_INI.open('rb') as handle:
                folder.create_file_from_filelike(handle, 'log.ini')

        return calcinfo



# -*- coding: utf-8 -*-
"""Parser for a :class:`aiida_pyscf.calculations.base.PyscfCalculation` job."""
from __future__ import annotations

import json
import pathlib

import dill
import numpy
from aiida.engine import ExitCode
from aiida.orm import ArrayData, Dict, SinglefileData, TrajectoryData
from aiida.parsers.parser import Parser
from aiida_shell.data import PickledData
from ase.io.extxyz import read_extxyz
from pint import UnitRegistry

from aiida_pyscf.calculations.base import PyscfCalculation

ureg = UnitRegistry()


class PyscfParser(Parser):
    """Parser for a :class:`aiida_pyscf.calculations.base.PyscfCalculation` job."""

    def __init__(self, *args, **kwargs):
        """Construct a new instance."""
        self.dirpath_temporary: pathlib.Path | None = None
        super().__init__(*args, **kwargs)

    def parse(self, retrieved_temporary_folder: str | None = None, **kwargs):  # noqa: PLR0912, PLR0915
        """Parse the contents of the output files stored in the ``retrieved`` output node.

        :returns: An exit code if the job failed.
        """
        self.dirpath_temporary = pathlib.Path(retrieved_temporary_folder) if retrieved_temporary_folder else None

        if 'parameters' in self.node.inputs:
            parameters = self.node.inputs.parameters.get_dict()
        else:
            parameters = {}

        try:
            with self.retrieved.base.repository.open(PyscfCalculation.FILENAME_STDOUT, 'r') as handle:
                handle.read()
        except FileNotFoundError:
            return self.handle_failure('ERROR_OUTPUT_STDOUT_MISSING')

        try:
            with self.retrieved.base.repository.open(PyscfCalculation.FILENAME_RESULTS, 'rb') as handle:
                parsed_json = json.load(handle)
        except FileNotFoundError:
            return self.handle_failure('ERROR_OUTPUT_RESULTS_MISSING')

        try:
            with self.retrieved.base.repository.open(PyscfCalculation.FILENAME_MODEL, 'rb') as handle:
                model = dill.load(handle)
        except FileNotFoundError:
            if parameters.get('results', {}).get('pickle_model', True):
                self.logger.warning(f'The pickled model file `{PyscfCalculation.FILENAME_MODEL}` could not be read.')
        except dill.UnpicklingError:
            self.logger.warning(f'The pickled model file `{PyscfCalculation.FILENAME_MODEL}` could not be unpickled.')
        else:
            self.out('model', PickledData(model))

        results_mean_field = parsed_json.setdefault('mean_field', {})

        if 'optimized_coordinates' in parsed_json.get('optimizer', {}):
            structure = self.node.inputs.structure.clone()
            optimized_coordinates = parsed_json['optimizer'].pop('optimized_coordinates') * ureg.bohr
            structure.reset_sites_positions(optimized_coordinates.to(ureg.angstrom).magnitude.tolist())
            self.out('structure', structure)

        if 'total_energy' in results_mean_field:
            energy = results_mean_field['total_energy'] * ureg.hartree
            results_mean_field['total_energy'] = energy.to(ureg.electron_volt).magnitude
            results_mean_field['total_energy_units'] = 'eV'

        if 'molecular_orbitals' in results_mean_field:
            labels = results_mean_field['molecular_orbitals']['labels']
            energies = results_mean_field['molecular_orbitals']['energies'] * ureg.hartree
            results_mean_field['molecular_orbitals']['energies'] = energies.to(ureg.electron_volt).magnitude.tolist()
            results_mean_field['molecular_orbitals']['labels'] = [label.strip() for label in labels]

        if 'forces' in results_mean_field:
            forces = results_mean_field['forces'] * ureg.hartree / ureg.bohr
            results_mean_field['forces'] = forces.to(ureg.electron_volt / ureg.angstrom).magnitude.tolist()
            results_mean_field['forces_units'] = 'eV/Å'

        if self.dirpath_temporary:
            for filepath_cubegen in self.dirpath_temporary.glob('mo_*.cube'):
                self.out(f'cubegen.orbitals.{filepath_cubegen.stem}', SinglefileData(filepath_cubegen))

            for filepath_density in self.dirpath_temporary.glob('density.cube'):
                self.out('cubegen.density', SinglefileData(filepath_density))

            for filepath_mep in self.dirpath_temporary.glob('mep.cube'):
                self.out('cubegen.mep', SinglefileData(filepath_mep))

            for filepath_fcidump in self.dirpath_temporary.glob('*.fcidump'):
                self.out(f'fcidump.{filepath_fcidump.stem}', SinglefileData(filepath_fcidump))

            for filepath_hessian in self.dirpath_temporary.glob('hessian.npy'):
                with filepath_hessian.open('rb') as handle:
                    hessian = numpy.load(handle)
                    array = ArrayData()
                    array.set_array('hessian', hessian)
                    self.out('hessian', array)

            filepath_trajectory = list(self.dirpath_temporary.glob('*_optim.xyz'))
            if filepath_trajectory:
                # There should be only one file that matches the pattern in the glob.
                trajectory = self.build_output_trajectory(filepath_trajectory[0])
                self.out('trajectory', trajectory)

        self.out('parameters', Dict(parsed_json))

        if results_mean_field['is_converged'] is False:
            return self.handle_failure('ERROR_ELECTRONIC_CONVERGENCE_NOT_REACHED', override_scheduler=True)

        if 'optimizer' in parsed_json and parsed_json['optimizer']['is_converged'] is False:
            return self.handle_failure('ERROR_IONIC_CONVERGENCE_NOT_REACHED', override_scheduler=True)

        return ExitCode(0)

    def build_output_trajectory(self, filepath_trajectory: pathlib.Path) -> TrajectoryData:
        """Build the ``TrajectoryData`` output node from the optimization file.

        :param filepath_trajectory: The filepath to the file containing the trajectory frames in XYZ format.
        :returns: The trajectory output node.
        """
        positions = []
        energies = []

        def batch(iterable, batch_size):
            """Split an iterable into a list of elements which size ``batch_size."""
            return [iterable[i : i + batch_size] for i in range(0, len(iterable), batch_size)]

        with filepath_trajectory.open() as handle:
            for atoms in read_extxyz(handle, index=slice(None, None)):
                positions.append(atoms.positions.tolist())
                # The ``atoms.info`` contains the parsed contents of the comment line in the XYZ frame. The geometry
                # optimizer does not use proper extended XYZ format, which is parsed by ASE as:
                #   {'Iteration': True, '0': True, 'Energy': True, '-74.96440482': True}
                # We convert this into a dictionary to extract the energy.
                properties = dict(batch(list(atoms.info), 2))
                energies.append(float(properties['Energy']))

        # Since the geometry optimization only supports atomic relaxation, the cell is fixed. The cells and the arrays
        # of symbols are therefore static and the input cell and symbol list can be copied for the number of frames.
        nframes = len(positions)
        trajectory = TrajectoryData()
        trajectory.set_trajectory(
            stepids=numpy.arange(nframes),
            cells=numpy.array([self.node.inputs.structure.cell] * nframes),
            symbols=[site.kind_name for site in self.node.inputs.structure.sites],
            positions=numpy.array(positions),
        )
        energies = (numpy.array(energies) * ureg.hartree).to(ureg.electron_volt).magnitude  # type: ignore[attr-defined]
        trajectory.set_array('energies', energies)

        return trajectory

    def handle_failure(self, exit_code_label: str, override_scheduler: bool = False) -> ExitCode:
        """Return the exit code corresponding to the given label unless the scheduler.

        This method also takes care of attaching the checkfile as an output if it was retrieved.

        :param override_scheduler: If set to ``True``, will return the given exit code even if one had already been set
            by the scheduler plugin.
        :returns: The exit code that should be returned by the caller.
        """
        self.attach_checkpoint_output()

        # If ``override_scheduler`` is ``False`` and an exit status has already been set by the scheduler, keep that by
        # returning it as is.
        if not override_scheduler and self.node.exit_status is not None:
            return ExitCode(self.node.exit_status, self.node.exit_message)

        # Either the scheduler parser did not return an exit code or we should override it regardless.
        return getattr(self.exit_codes, exit_code_label)

    def attach_checkpoint_output(self) -> None:
        """Attach the checkpoint file as an output if it was retrieved in the temporary folder."""
        if self.dirpath_temporary is None:
            return

        self.out('checkpoint', SinglefileData(self.dirpath_temporary / PyscfCalculation.FILENAME_CHECKPOINT))
